==> w9_pathfinding-0.0.1.tar.gz: w9_pathfinding-0.0.1/w9_pathfinding/src/include/multi_agent_a_star.h <==
#pragma once

#include "core.h"
#include "resumable_search.h"
#include "space_time_a_star.h"

namespace maas {

    class MultiAgentAStar : public AbsMAPF {
        // A* with Operator Decomposition
        // Standley, T.S.: Finding optimal solutions to cooperative pathfinding problems.
        // In: AAAI Conference on Artificial Intelligence. pp. 173–178 (2010)

        public:
            AbsGraph* graph;
            MultiAgentAStar(AbsGraph* graph);

            vector<Path> mapf(vector<int> starts, vector<int> goals);
            vector<Path> mapf(
                vector<int> starts,
                vector<int> goals,
                int max_length,
                double max_time,
                bool operator_decomposition,
                const ReservationTable *rt
            );
    };

    struct Agent {
        int start, goal;
        ResumableSearch *rrs;

        Agent(int start, int goal, ResumableSearch* rrs) : start(start), goal(goal), rrs(rrs) {};
        ~Agent() {delete rrs;};
    };

    struct Node {
        int id, parent, time;
        double distance;
        vector<int> positions;
        
        Node(int id, int parent, int time, double distance, vector<int>& positions) :
            id(id),
            parent(parent),
            time(time),
            distance(distance),
            positions(positions) {};
    };

    typedef vector<Node> Tree;
    typedef pair<double, int> key;
    typedef priority_queue<key, vector<key>, std::greater<key>> Queue;

    class AStarSolver {
        public:
            AStarSolver(AbsGraph* graph, vector<int>& starts, vector<int>& goals, const ReservationTable* rt);

            vector<Path> find_paths(int max_length, double max_time);

        protected:
            AbsGraph* graph_;
            const ReservationTable* rt_;
            Tree tree_;
            vector<Agent> agents_;
            int num_agents_;

            void print_node(const Node& node) const;

            bool is_goal(vector<int>& positions);

            // returns a list of free adjacent tiles for each position and the cost of moving to that tile
            vector<vector<pair<int, double>>> get_neighbors(Node& node);

            // returns a path that ends with node_id
            vector<Path> reconstruct_paths(int node_id);

            // creates a new node and adds it to the tree
            Node create_node(int parent, int time, double distance, vector<int> &positions);

            // returns the amount of time that an agent has been at its goal position
            int get_waiting_time(int node_id, int agent_id);

            // the minimum total cost of paths from this position to the goal
            // returns -1 if at least one path is not possible
            double heuristic(vector<int>& positions);

            vector<int> to_hash(vector<int>& positions, int parent);
    };

    class AStarODSolver : AStarSolver {
        public:
            AStarODSolver(AbsGraph* graph, vector<int>& starts, vector<int>& goals, const ReservationTable* rt);

            vector<Path> find_paths(int max_length, double max_time);

        private:
            vector<pair<int, double>> get_neighbors(Node& node);
            std::unordered_set<int> get_occupied_nodes(Node& node);
    };

}

==> w9_pathfinding-0.1.0.tar.gz: w9_pathfinding-0.1.0/w9_pathfinding/src/include/multi_agent_a_star.h <==
#pragma once

#include <memory>
#include "core.h"
#include "resumable_search.h"
#include "space_time_a_star.h"

namespace maas {

    class MultiAgentAStar : public AbsMAPF {
        // A* with Operator Decomposition
        // Standley, T.S.: Finding optimal solutions to cooperative pathfinding problems.
        // In: AAAI Conference on Artificial Intelligence. pp. 173–178 (2010)

        public:
            AbsGraph* graph;
            MultiAgentAStar(AbsGraph* graph);

            vector<Path> mapf(vector<int> starts, vector<int> goals);
            vector<Path> mapf(
                vector<int> starts,
                vector<int> goals,
                int max_length,
                double max_time,
                bool operator_decomposition,
                const ReservationTable *rt
            );
    };

    struct Agent {
        int start, goal;
        std::unique_ptr<ResumableSearch> rrs;

        Agent(int start, int goal, std::unique_ptr<ResumableSearch> rrs)
            : start(start), goal(goal), rrs(std::move(rrs)) {}
    };

    struct Node {
        int id, parent, time;
        double distance;
        vector<int> positions;
        
        Node(int id, int parent, int time, double distance, vector<int>& positions) :
            id(id),
            parent(parent),
            time(time),
            distance(distance),
            positions(positions) {};
    };

    typedef vector<Node> Tree;
    typedef pair<double, int> key;
    typedef priority_queue<key, vector<key>, std::greater<key>> Queue;

    class AStarSolver {
        public:
            AStarSolver(AbsGraph* graph, vector<int>& starts, vector<int>& goals, const ReservationTable* rt);

            vector<Path> find_paths(int max_length, double max_time);

        protected:
            AbsGraph* graph_;
            const ReservationTable* rt_;
            Tree tree_;
            vector<Agent> agents_;
            int num_agents_;

            void print_node(const Node& node) const;

            bool is_goal(vector<int>& positions);

            // returns a list of free adjacent tiles for each position and the cost of moving to that tile
            vector<vector<pair<int, double>>> get_neighbors(Node& node);

            // returns a path that ends with node_id
            vector<Path> reconstruct_paths(int node_id);

            // creates a new node and adds it to the tree
            Node create_node(int parent, int time, double distance, vector<int> &positions);

            // returns the amount of time that an agent has been at its goal position
            int get_waiting_time(int node_id, int agent_id);

            // the minimum total cost of paths from this position to the goal
            // returns -1 if at least one path is not possible
            double heuristic(vector<int>& positions);

            vector<int> to_hash(vector<int>& positions, int parent);
    };

    class AStarODSolver : AStarSolver {
        public:
            AStarODSolver(AbsGraph* graph, vector<int>& starts, vector<int>& goals, const ReservationTable* rt);

            vector<Path> find_paths(int max_length, double max_time);

        private:
            vector<pair<int, double>> get_neighbors(Node& node);
            std::unordered_set<int> get_occupied_nodes(Node& node);
    };

}
